/*!
 * Copyright (C) 2001-2003 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

#include <dev/board.h>
#include <dev/debug.h>
#include <dev/urom.h>

#include <sys/version.h>
#include <sys/heap.h>
#include <sys/timer.h>
#include <sys/device.h>

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

#include <hxmp3/mp3dec.h>

#include "at73dac.h"

#define MAX_FILESIZE    (1024 * 1024)

HMP3Decoder hdec;
MP3FrameInfo frameinfo;
short pcmdec[4096];
short pcmbuf[4096];

/*
 * Play MP3 file from local file system.
 *
 * \param path Pathname of the MP3 file to play.
 *
 * \return 0 on success, -1 if opening the file failed.
 */
static int PlayMp3File(char *path)
{
    int fd;
    u_char *mp3buf;
    u_char *mp3ptr;
    int got;
    int skip;
    int rbytes;
    int ec;
    int first = 1;
    int i;
    short last[2] = { 0, 0 };
    short diff;

    /*
     * Open the MP3 file.
     */
    printf("Play %s: ", path);
    if ((fd = _open(path, _O_RDONLY | _O_BINARY)) == -1) {
        printf("Error %d\n", errno);
        return -1;
    }
    puts("OK");

    mp3buf = malloc(MAX_FILESIZE);
    got = _read(fd, mp3buf, MAX_FILESIZE);
    printf("[%d]", got);
    _close(fd);

    mp3ptr = mp3buf;
    rbytes = got;
    for (;;) {
        if (rbytes <= 0) {
            At73DacFlush();
            break;
        }
        if ((skip = MP3FindSyncWord(mp3ptr, rbytes)) < 0) {
            printf("\n[ERR-SYN]");
            break;
        }
        if (skip) {
            printf("[Skip %d]", skip);
            mp3ptr += skip;
            rbytes -= skip;
            if (rbytes <= 0) {
                At73DacFlush();
                break;
            }
        }
        ec = MP3Decode(hdec, &mp3ptr, &rbytes, pcmdec, 0);
        if (ec) {
            printf("\n[ERR-DEC]");
            break;
        }
        MP3GetLastFrameInfo(hdec, &frameinfo);
        if (first) {
            first = 0;
            printf("\nLayer %d, ", frameinfo.layer);
            printf("Version %d, ", frameinfo.version);
            printf("Bitrate %d, ", frameinfo.bitrate);
            printf("Channels %d, ", frameinfo.nChans);
            printf("Sample rate %d, ", frameinfo.samprate);
            printf("Bits/sample %d\n", frameinfo.bitsPerSample);
        }
        printf("[%d]", frameinfo.outputSamps);
#if 1
        for (i = 0; i < frameinfo.outputSamps; i += 2) {
            diff = (pcmdec[i] - last[0]) >> 2;
            last[0] += diff;
            pcmbuf[i * 4 + 0] = last[0];
            last[0] += diff;
            pcmbuf[i * 4 + 2] = last[0];
            last[0] += diff;
            pcmbuf[i * 4 + 4] = last[0];
            last[0] += diff;
            pcmbuf[i * 4 + 6] = last[0];

            diff = (pcmdec[i + 1] - last[1]) >> 2;
            last[1] += diff;
            pcmbuf[i * 4 + 1] = last[1];
            last[1] += diff;
            pcmbuf[i * 4 + 3] = last[1];
            last[1] += diff;
            pcmbuf[i * 4 + 5] = last[1];
            last[1] += diff;
            pcmbuf[i * 4 + 7] = last[1];
        }
        if (At73DacWrite(pcmbuf, frameinfo.outputSamps * 4)) {
            printf("\n[ERR-ADC]");
        }
#else
        if (At73DacWrite(pcmdec, frameinfo.outputSamps)) {
            printf("\n[ERR-ADC]");
        }
#endif
        //printf("[%d]", rbytes);
        //mp3ptr += rbytes;
        //got -= rbytes;
    }


    free(mp3buf);
    printf("\nDone, %u bytes free\n", (u_int)NutHeapAvailable());
    return 0;
}

/*!
 * \brief Main application routine. 
 *
 */
int main(void)
{
    /* Baudrate for debug output. */
    u_long baud = 115200;

    /*
     * Assign stdout to the debug device.
     */
    NutRegisterDevice(&DEV_DEBUG, 0, 0);
    freopen(DEV_DEBUG_NAME, "w", stdout);
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    printf("\n\nMP3 player on Nut/OS %s\n", NutVersionString());

    /*
     * Initialize file system.
     */
    NutRegisterDevice(&devUrom, 0, 0);

    NutSleep(200);
    if (At73DacAdaptMasterClock(44100)) {
        printf("Failed to set DAC clocks.\n");
        for (;;);
    }
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    printf("CPU Clock   : %lu\n", NutGetCpuClock());
    printf("Master Clock: %lu\n", At91GetMasterClock());
    printf("%lu system ticks/s\n", NutGetTickClock());
    NutSleep(2000);

    hdec = MP3InitDecoder();
    At73DacInit();

    for (;;) {
        PlayMp3File("UROM:xmas.mp3");
        NutSleep(1000);
    }


    for (;;);
}
